//#################################################################################################//
//#                                                                                               #// 
//#  **** Schneider Engineering, Bad Vilbel ***                                                   #//
//#  **** Inh.  Dipl.-Ing. Peter Schneider  ***                                                   #//
//#                                                                                               #//
//#  Filename : bug2Library.h                                                                     #//
//#  Project  : Odemetrie                                                                         #//
//#                                                                                               #//
//#                                                                                               #//    
//#  Odemetrie mit bug2 Pathfindig Algorithmus ein moeglicher Implementierungsvorschlag           #// 
//#  Quelle des Algorithmus www.wikipedia.de                                                      #//
//#                                                                                               #//
//#  (c) Peter Schneider (Schneider Engineering software / hardware development)                  #//
//#  www.ps-robotics.com                                                                          #//
//#                                                                                               #//
//#  Version:                                                                                     #//         
//#  1.03       14 July 2011 11:55       Original release                                         #//
//#                                                                                               #//
//# This program is free software; you can redistribute it and/or modify                          #//
//# it under the terms of the GNU General Public License as published by                          #//
//# the Free Software Foundation; either version 2 of the License, or                             #//
//# (at your option) any later version.                                                           #//
//#                                                                                               #//		
//# This program is distributed in the hope that it will be useful,                               #//
//# but WITHOUT ANY WARRANTY; without even the implied warranty of                                #//	
//# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                 #//	
//# GNU General Public License for more details.                                                  #//
//#                                                                                               #//
//# You should have received a copy of the GNU General Public License                             #//
//# along with this program; if not, write to peter.schneider@ps-robotics.de                      #//
//#                                                                                               #//
//#################################################################################################//
//
//   Erkennen des Hindernis via Sensoren
//  
//               H            X entspricht Position des Roboters
//               |             H enspricht Hindernissen die mit Sensoren erfasst werden koennen
//              X - H
//               | 
//               H
// 
//   Moegliche Moverichtungen in  Move_Richtung_X  und  Move_Richtung_Y  gespeichert
//              
//              LEFT
//               |
//       BWD  -- X --  FWD   
//               |
//             RIGHT
//                       
// Grundstzliches Vorgehen bei BUG 1 und BUG 2 Algorithmus:
// -  Bewege dich gerade auf einer Linie Richtung Ziel
// -  Bewege dich nicht diagonal zur X-Y-Achse
// -  Folge der Wand (per default nach rechts)
//
// Algorithmus-Beschreibung:
// 1)	Folge der m-Line in Richtung des Ziels. Bewege dich gerade auf einer Linie Richtung Ziel
// 2)	Wenn ein Hindernis im Weg ist, umfahre das Hindernis (Folge der Wand (links oder rechts)) bis Du wieder auf die m-Line triffst, welcher nher zum Ziel ist 
// 3)	Verlasse das Hindernis und folge der m-Line in Richtung Ziel  
//
//
// **** benoetigte Header Dateien ****
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <ctype.h>


// Laden der Defaultwerte
#include "bug2Defaults_v3.h"
#include "help_routines_v3.h"



// ***********  Globale Funktionen ***************
// *****
void  init_hindernis (void);
void  ZeichneWeg (void);
void  rahmenzeichnen (int , int , int , int );
void  Coursor_Zeichnen (int , int );
void  matrix_anzeigen (void);
void  Check_For_Hindernis (void);





// -----------------------------------------------------------------------------------------------
// *************  Pruefen ob ein Hindernis auf dem Weg des Roboters ist ************* 
// *****
void  Check_For_Hindernis (void)
{
   sensor_mitte  = 0;
   sensor_links  = 0;
   sensor_rechts = 0;
   
   
   if (move_dir == FWD)               // Diese Berechnung ist bei realen Sensoren durch eine Entfernungsmessung zum
   {                                // Hindernis zu ersetzen. Aussage dann z.B.  0 - 1024 ueber AD-Wandler
      if (matrix [X+1] [Y] == 1)    // move_dir = 3  bedeutet bewegungsrichtung forwaerts
          sensor_mitte  = 1;        // dementsprechend mssen die Felder der Matrix abgefragt werden kann bei realen 
                                    // Sensoren entfallen. Nur fr die Simulation am PC ntig
      if (matrix [X] [Y-1] == 1)
          sensor_links  = 1;
             
      if (matrix [X] [Y+1] == 1)
          sensor_rechts  = 1;
   }
   
   
   else if (move_dir == BWD)
   { 
      if (matrix [X-1] [Y] == 1)    // move_dir = 4 / BWD  bedeutet Bewegungsrichtung rueckwaerts
          sensor_mitte  = 1;
     
      if (matrix [X] [Y+1] == 1)
          sensor_links  = 1;
          
      if (matrix [X] [Y-1] == 1) 
          sensor_rechts  = 1;      
   }
   
   
   else if (move_dir == LEFT)
   { 
      if (matrix [X] [Y-1] == 1)    // move_dir = 1 / LEFT  bedeutet bewegungsrichtung links
          sensor_mitte  = 1;
     
      if (matrix [X-1] [Y] == 1) 
          sensor_links  = 1;
          
      if (matrix [X+1] [Y] == 1)
          sensor_rechts  = 1;       
   }    
    
   else if (move_dir == RIGHT)
   {
      if (matrix [X] [Y+1] == 1)    // move_dir = 2 / RIGHT  bedeutet bewegungsrichtung rechts
          sensor_mitte  = 1;
     
      if (matrix [X+1] [Y] == 1) 
          sensor_links  = 1;
          
      if (matrix [X-1] [Y] == 1)
          sensor_rechts  = 1;               
   }      
     
   // Datenbank Query vorbereiten Einzelabfragen in eine fuenfstellige Zahl umwandel
   db_query = sensor_links * 10000 + sensor_mitte * 1000 + sensor_rechts * 100 + move_dir * 10 + last_dir;

     
}
// -----------------------------------------------------------------------------------------------












// -----------------------------------------------------------------------------------------------
// *************  Hindenis umfahren mit BUG2 Algorithmus ************* 
//
// Algorithmus-Beschreibung:
// 1)	Folge der m-Line in Richtung des Ziels. Bewege dich gerade auf einer Linie Richtung Ziel
// 2)	Wenn ein Hindernis im Weg ist, umfahre das Hindernis (Folge der Wand (links oder rechts)) bis Du wieder auf die m-Line triffst, welcher nher zum Ziel ist 
// 3)	Verlasse das Hindernis und folge der m-Line in Richtung Ziel  
// *****
void  bug2_algo (int startX, int startY, int zielX, int zielY) 
{  
    
   int   hitpoint_X;
   int   c;                                     
   int   k;
   int   i;
   int   temp_db;
   
   X = hitpoint_X = startX;         // hitpoint = merken der Position wo der Algorithmus aufgerufen wurde
   Y = startY;
   
   
   // Start Ausweichen mit BUG2 Algorithmus
   while (1)
   {
     
      
      Coursor_Zeichnen (X , Y);
      Sleep (200);
      
      Check_For_Hindernis ();                   // Sensoren ueberprfen 
      

    
      //--------------------------------------------------------------------------------------
      // Ziel erreicht Algorithmus verlassen                                   
      if ((Y == zielY) && (X == zielX))         
      {
         Coursor_Zeichnen (X , Y);

         gotoxy (46, 25);
         printf ("F E R T I G");        
         
         break; 
      } 
  
            

      //--------------------------------------------------------------------------------------
      // Einschwenken auf die Zielgerade wenn Sensor Rechts den Weg als frei zeigt
      // und der jetztige Position nher zum Ziel ist "hitpoint < X
      if ((Y == zielY)  && (sensor_rechts == 0) && (move_dir == 1) && (hitpoint_X < X))  
      {
         move_dir = 3; 
         last_dir = 3;    
         X = X + 1;
         break; 
      }              
      
      // Einschwenken auf die Zielgerade wenn Sensor Links den Weg als frei zeigt   
      else if ((Y == zielY)  && (sensor_links == 0)&& (move_dir == 2) && (hitpoint_X < X)) 
      {
         move_dir = 3;  
         last_dir = 3;    
         X = X + 1;       
         break;
      }                                                
      //--------------------------------------------------------------------------------------


         
      
      
      
 
      //-------------------------------------------------------------------------------------- 
      // An dem Hindernis vorbeigehen und Kontakt zur Wand / Hindernis halten Sensor Links True RICHTUNG von Ziellinie weg 
      
      for (i = 0; i < 25; i++)   // gewichtsmatrix auswerten
      {     
         temp_db = gewicht[i][0] * 10000 + gewicht[i][1] * 1000 + gewicht[i][2] * 100  + gewicht[i][3] * 10 + gewicht[i][4];       
           
         if ( db_query == temp_db )
         {    
            X        = X + (gewicht[i][5] - 5);     // Die 5 wieder abziehen so das X / Y
            Y        = Y + (gewicht[i][6] - 5);     // wieder -1 / 0 / 1  wird
            move_dir = gewicht[i][7];
            last_dir = gewicht[i][8];
           
            gotoxy (46, 19);
            printf ("                                            ");
            gotoxy (46, 19);
            printf ("X=%d  Y=%d  mv_dir=%d  last=%d", X, Y, move_dir, last_dir);
            
            Coursor_Zeichnen (X , Y);
         }
      
      
      }
      
      
                                                   
      Coursor_Zeichnen (X , Y);
  
        
     
   
       
   }   // Ende While

}
// -----------------------------------------------------------------------------------------------





